@@ -138,6 +138,7 @@ class POP3initMixIn(object):
         self.log.trace('POP3 connection %s established' % self.conn
                        + os.linesep)
 
+
 #######################################
 class Py24POP3SSLinitMixIn(object):
     '''Mix-In class to do POP3 over SSL initialization with Python 2.4's
@@ -184,6 +185,7 @@ class Py24POP3SSLinitMixIn(object):
         self.log.trace('POP3 connection %s established' % self.conn
                        + os.linesep)
 
+
 #######################################
 class Py23POP3SSLinitMixIn(object):
     '''Mix-In class to do POP3 over SSL initialization with custom-implemented
@@ -230,6 +232,7 @@ class Py23POP3SSLinitMixIn(object):
         self.log.trace('POP3 SSL connection %s established' % self.conn
                        + os.linesep)
 
+
 #######################################
 class IMAPinitMixIn(object):
     '''Mix-In class to do IMAP non-SSL initialization.
@@ -327,7 +330,7 @@ class RetrieverSkeleton(ConfigurableBase):
                      - name - parameter name
                      - type - a type function to compare the parameter value
                      against (i.e. str, int, bool)
-                     - default - optional default value.  If not preseent, the
+                     - default - optional default value.  If not present, the
                      parameter is required.
 
       __str__(self) - return a simple string representing the class instance.
@@ -378,6 +381,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self.__initialized = False
         self.gotmsglist = False
         self._clear_state()
+        self.conn = None
         ConfigurableBase.__init__(self, **args)
 
     def _clear_state(self):
@@ -429,15 +433,16 @@ class RetrieverSkeleton(ConfigurableBase):
                 or (isinstance(mailbox, (str, unicode)) and mailbox)), (
             'bad mailbox %s (%s)' % (mailbox, type(mailbox))
         )
-        if isinstance(mailbox, str):
-            mailbox = mailbox.decode('utf-8')
         filename = self.oldmail_filename
-        if mailbox is None:
-            # No mailbox (POP), use above with no extension
-            pass
-        else:
+        if mailbox is not None:
+            if isinstance(mailbox, str):
+                mailbox = mailbox.decode('utf-8')
+            mailbox = re.sub(STRIP_CHAR_RE, '.', mailbox)
+            mailbox = mailbox.encode('utf-8')
             # Use oldmail file per IMAP folder
-            filename += '-' + re.sub(STRIP_CHAR_RE, '.', mailbox)
+            filename += '-' + mailbox
+        # else:
+            # mailbox is None, is POP, just use filename
         return filename
 
     def oldmail_exists(self, mailbox):
@@ -469,6 +474,13 @@ class RetrieverSkeleton(ConfigurableBase):
                 continue
             try:
                 (msgid, timestamp) = line.split('\0', 1)
+                if msgid.count('/') == 2:
+                    # Was pre-4.22.0 file format, which includes the
+                    # mailbox name in the msgid, in the format
+                    # 'uidvalidity/mailbox/serveruid'.
+                    # Strip it out.
+                    fields = msgid.split('/')
+                    msgid = '/'.join([fields[0], fields[2]])
                 self.oldmail[msgid] = int(timestamp)
             except ValueError:
                 # malformed
@@ -579,6 +591,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self._delmsgbyid(msgid)
         self.deleted[msgid] = True
 
+
 #######################################
 class POP3RetrieverBase(RetrieverSkeleton):
     '''Base class for single-user POP3 mailboxes.
@@ -617,6 +630,10 @@ class POP3RetrieverBase(RetrieverSkeleton):
             for (i, line) in enumerate(msglist):
                 try:
                     (msgnum, msgid) = line.split(None, 1)
+                    # Don't allow / in UIDs we store, as we look for that to 
+                    # detect old-style oldmail files.  Shouldn't occur in POP3
+                    # anyway.
+                    msgid = msgid.replace('/', '-')
                 except ValueError:
                     # Line didn't contain two tokens.  Server is broken.
                     raise getmailOperationError(
@@ -712,8 +729,8 @@ class POP3RetrieverBase(RetrieverSkeleton):
         # Handle password
         if self.conf.get('password', None) is None:
             self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'], 'pop3', 
-                self.log
+                self, self.conf['username'], self.conf['server'],
+                self.received_with, self.log
             )
         RetrieverSkeleton.initialize(self, options)
         try:
@@ -739,25 +756,28 @@ class POP3RetrieverBase(RetrieverSkeleton):
     def abort(self):
         self.log.trace()
         RetrieverSkeleton.abort(self)
+        if not self.conn:
+            return
         try:
             self.conn.rset()
             self.conn.quit()
         except (poplib.error_proto, socket.error), o:
             pass
-        del self.conn
+        self.conn = None
 
     def quit(self):
         RetrieverSkeleton.quit(self)
         self.log.trace()
-        if not getattr(self, 'conn', None):
+        if not self.conn:
             return
         try:
             self.conn.quit()
-            self.conn = None
         except (poplib.error_proto, socket.error), o:
             raise getmailOperationError('POP error (%s)' % o)
         except AttributeError:
             pass
+        self.conn = None
+
 
 #######################################
 class MultidropPOP3RetrieverBase(POP3RetrieverBase):
@@ -819,6 +839,7 @@ class MultidropPOP3RetrieverBase(POP3RetrieverBase):
         msg.recipient = address_no_brackets(line.strip())
         return msg
 
+
 #######################################
 class IMAPRetrieverBase(RetrieverSkeleton):
     '''Base class for single-user IMAP mailboxes.
@@ -999,7 +1020,8 @@ class IMAPRetrieverBase(RetrieverSkeleton):
 
         self.log.debug('selecting mailbox "%s"' % mailbox + os.linesep)
         try:
-            if self.app_options['delete'] or self.app_options['delete_after']:
+            if (self.app_options['delete'] or self.app_options['delete_after'] 
+                    or self.app_options['delete_bigger_than']):
                 read_only = False
             else:
                 read_only = True
@@ -1039,9 +1061,11 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 )
                 for line in response:
                     r = self._parse_imapattrresponse(line)
-                    msgid = (
-                        '%s/%s' % (self.uidvalidity, r['uid'])
-                    )
+                    # Don't allow / in UIDs we store, as we look for that to 
+                    # detect old-style oldmail files.  Can occur with IMAP, at 
+                    # least with some servers.
+                    uid = r['uid'].replace('/', '-')
+                    msgid = '%s/%s' % (self.uidvalidity, uid)
                     self._mboxuids[msgid] = r['uid']
                     self._mboxuidorder.append(msgid)
                     self.msgnum_by_msgid[msgid] = None
@@ -1128,18 +1152,77 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 # response[0] is None instead of a message tuple
                 raise getmailRetrievalError('failed to retrieve msgid %s' 
                                             % msgid)
+
+            # record mailbox retrieved from in a header
+            msg.add_header('X-getmail-retrieved-from-mailbox', 
+                           self.mailbox_selected)
+
+            # google extensions: apply labels, etc
+            if 'X-GM-EXT-1' in self.conn.capabilities:
+                metadata = self._getgmailmetadata(uid, msg)
+                for (header, value) in metadata.items():
+                    msg.add_header(header, value)
+
             return msg
 
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
 
+    def _getgmailmetadata(self, uid, msg):
+        """
+        Add Gmail labels and other metadata which Google exposes through an
+        IMAP extension to headers in the message.
+        
+        See https://developers.google.com/google-apps/gmail/imap_extensions
+        """
+        try:
+            # ['976 (X-GM-THRID 1410134259107225671 X-GM-MSGID '
+            #   '1410134259107225671 X-GM-LABELS (labels space '
+            #   'separated) UID 167669)']
+            response = self._parse_imapuidcmdresponse('FETCH', uid,
+                '(X-GM-LABELS X-GM-THRID X-GM-MSGID)')
+        except imaplib.IMAP4.error, o:
+            self.log.warning('Could not fetch google imap extensions: %s' % o)
+            return {}
+
+        if not response:
+            return {}
+            
+        ext = re.search(
+            'X-GM-THRID (?P<THRID>\d+) X-GM-MSGID (?P<MSGID>\d+)'
+            ' X-GM-LABELS \((?P<LABELS>.*)\) UID',
+            response[0]
+        )
+        if not ext:
+            self.log.warning(
+                'Could not parse google imap extensions. Server said: %s'
+                % repr(response)
+            )
+            return {}
+
+        results = ext.groupdict()
+        metadata = {}
+        for item in ('LABELS', 'THRID', 'MSGID'):
+            if item in results and results[item]:
+                metadata['X-GMAIL-%s' % item] = results[item]
+
+        return metadata
+
     def _getmsgbyid(self, msgid):
         self.log.trace()
-        return self._getmsgpartbyid(msgid, '(BODY.PEEK[])')
+        if self.conf.get('use_peek', True):
+            part = '(BODY.PEEK[])'
+        else:
+            part = '(RFC822)'
+        return self._getmsgpartbyid(msgid, part)
 
     def _getheaderbyid(self, msgid):
         self.log.trace()
-        return self._getmsgpartbyid(msgid, '(BODY.PEEK[header])')
+        if self.conf.get('use_peek', True):
+            part = '(BODY.PEEK[header])'
+        else:
+            part = '(RFC822[header])'
+        return self._getmsgpartbyid(msgid, part)
 
     def initialize(self, options):
         self.log.trace()
@@ -1148,8 +1231,8 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         if (self.conf.get('password', None) is None
                 and not (HAVE_KERBEROS_GSS and self.conf['use_kerberos'])):
             self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'], 'imap',
-                self.log
+                self, self.conf['username'], self.conf['server'], 
+                self.received_with, self.log
             )
             
         RetrieverSkeleton.initialize(self, options)
@@ -1203,14 +1286,17 @@ class IMAPRetrieverBase(RetrieverSkeleton):
     def abort(self):
         self.log.trace()
         RetrieverSkeleton.abort(self)
+        if not self.conn:
+            return
         try:
             self.quit()
         except (imaplib.IMAP4.error, socket.error), o:
             pass
+        self.conn = None
 
     def quit(self):
         self.log.trace()
-        if not getattr(self, 'conn', None):
+        if not self.conn:
             return
         try:
             if self.mailbox_selected is not False:
@@ -1221,11 +1307,12 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 self.conn.expunge()
                 self.conn.close()
             self.conn.logout()
-            self.conn = None
         except imaplib.IMAP4.error, o:
             #raise getmailOperationError('IMAP error (%s)' % o)
             self.log.warning('IMAP error during logout (%s)' % o + os.linesep)
         RetrieverSkeleton.quit(self)
+        self.conn = None
+
 
 #######################################
 class MultidropIMAPRetrieverBase(IMAPRetrieverBase):
